#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import subprocess
import threading
from pathlib import Path

from prometheus_client import Gauge

from workspaces.workflow.schema import WorkflowRequest

monitors_running = Gauge("wf_monitors_running", "Number of active wf_monitor processes managed by this server")


class WorkflowMonitor:
    """
    Wraps the subprocess call for launching a workflow monitor.
    """

    def __init__(self, log_file: Path, request: WorkflowRequest, is_dag=False):
        self.log_file = log_file
        self.request = request
        self.is_dag = is_dag

    def start(self):
        """
        Create and start the wf_monitor process
        """
        # create the wf_monitor process
        monitor_process = subprocess.Popen(
            ["wf_monitor", self.log_file, str(self.request.workflow_request_id), str(self.is_dag)],
            cwd=self.request.results_dir,
        )

        # now we can increment the counter of wf_monitors running
        monitors_running.inc()

        # now create a thread to watch the wf_monitor process to decrement the gauge
        threading.Thread(target=self.wait_for_completion, args=[monitor_process]).start()

    def wait_for_completion(self, monitor_process: subprocess.Popen):
        # all we really need to do now is wait for this damn thing to finish
        # a separate thread is required here only to ensure that we don't block
        # the rest of the workflow service waiting on this to happen
        monitor_process.wait()

        # now we can decrement the monitors_running metric
        monitors_running.dec()
